// This file is generated by Tools/cases_generator/opcode_id_generator.py
// from:
//   Python/bytecodes.c
// Do not edit!

#ifndef Py_OPCODE_IDS_H
#define Py_OPCODE_IDS_H
#ifdef __cplusplus
extern "C" {
#endif

/* Instruction opcodes for compiled code */
#define CACHE                                    0
#define BINARY_SLICE                             1
#define BUILD_TEMPLATE                           2
#define BINARY_OP_INPLACE_ADD_UNICODE            3
#define CALL_FUNCTION_EX                         4
#define CHECK_EG_MATCH                           5
#define CHECK_EXC_MATCH                          6
#define CLEANUP_THROW                            7
#define DELETE_SUBSCR                            8
#define END_FOR                                  9
#define END_SEND                                10
#define EXIT_INIT_CHECK                         11
#define FORMAT_SIMPLE                           12
#define FORMAT_WITH_SPEC                        13
#define GET_AITER                               14
#define GET_ANEXT                               15
#define GET_ITER                                16
#define RESERVED                                17
#define GET_LEN                                 18
#define GET_YIELD_FROM_ITER                     19
#define INTERPRETER_EXIT                        20
#define LOAD_BUILD_CLASS                        21
#define LOAD_LOCALS                             22
#define MAKE_FUNCTION                           23
#define MATCH_KEYS                              24
#define MATCH_MAPPING                           25
#define MATCH_SEQUENCE                          26
#define NOP                                     27
#define NOT_TAKEN                               28
#define POP_EXCEPT                              29
#define POP_ITER                                30
#define POP_TOP                                 31
#define PUSH_EXC_INFO                           32
#define PUSH_NULL                               33
#define RETURN_GENERATOR                        34
#define RETURN_VALUE                            35
#define SETUP_ANNOTATIONS                       36
#define STORE_SLICE                             37
#define STORE_SUBSCR                            38
#define TO_BOOL                                 39
#define UNARY_INVERT                            40
#define UNARY_NEGATIVE                          41
#define UNARY_NOT                               42
#define WITH_EXCEPT_START                       43
#define BINARY_OP                               44
#define BUILD_INTERPOLATION                     45
#define BUILD_LIST                              46
#define BUILD_MAP                               47
#define BUILD_SET                               48
#define BUILD_SLICE                             49
#define BUILD_STRING                            50
#define BUILD_TUPLE                             51
#define CALL                                    52
#define CALL_INTRINSIC_1                        53
#define CALL_INTRINSIC_2                        54
#define CALL_KW                                 55
#define COMPARE_OP                              56
#define CONTAINS_OP                             57
#define CONVERT_VALUE                           58
#define COPY                                    59
#define COPY_FREE_VARS                          60
#define DELETE_ATTR                             61
#define DELETE_DEREF                            62
#define DELETE_FAST                             63
#define DELETE_GLOBAL                           64
#define DELETE_NAME                             65
#define DICT_MERGE                              66
#define DICT_UPDATE                             67
#define END_ASYNC_FOR                           68
#define EXTENDED_ARG                            69
#define FOR_ITER                                70
#define GET_AWAITABLE                           71
#define IMPORT_FROM                             72
#define IMPORT_NAME                             73
#define IS_OP                                   74
#define JUMP_BACKWARD                           75
#define JUMP_BACKWARD_NO_INTERRUPT              76
#define JUMP_FORWARD                            77
#define LIST_APPEND                             78
#define LIST_EXTEND                             79
#define LOAD_ATTR                               80
#define LOAD_COMMON_CONSTANT                    81
#define LOAD_CONST                              82
#define LOAD_DEREF                              83
#define LOAD_FAST                               84
#define LOAD_FAST_AND_CLEAR                     85
#define LOAD_FAST_BORROW                        86
#define LOAD_FAST_BORROW_LOAD_FAST_BORROW       87
#define LOAD_FAST_CHECK                         88
#define LOAD_FAST_LOAD_FAST                     89
#define LOAD_FROM_DICT_OR_DEREF                 90
#define LOAD_FROM_DICT_OR_GLOBALS               91
#define LOAD_GLOBAL                             92
#define LOAD_NAME                               93
#define LOAD_SMALL_INT                          94
#define LOAD_SPECIAL                            95
#define LOAD_SUPER_ATTR                         96
#define MAKE_CELL                               97
#define MAP_ADD                                 98
#define MATCH_CLASS                             99
#define POP_JUMP_IF_FALSE                      100
#define POP_JUMP_IF_NONE                       101
#define POP_JUMP_IF_NOT_NONE                   102
#define POP_JUMP_IF_TRUE                       103
#define RAISE_VARARGS                          104
#define RERAISE                                105
#define SEND                                   106
#define SET_ADD                                107
#define SET_FUNCTION_ATTRIBUTE                 108
#define SET_UPDATE                             109
#define STORE_ATTR                             110
#define STORE_DEREF                            111
#define STORE_FAST                             112
#define STORE_FAST_LOAD_FAST                   113
#define STORE_FAST_STORE_FAST                  114
#define STORE_GLOBAL                           115
#define STORE_NAME                             116
#define SWAP                                   117
#define UNPACK_EX                              118
#define UNPACK_SEQUENCE                        119
#define YIELD_VALUE                            120
#define RESUME                                 128
#define BINARY_OP_ADD_FLOAT                    129
#define BINARY_OP_ADD_INT                      130
#define BINARY_OP_ADD_UNICODE                  131
#define BINARY_OP_EXTEND                       132
#define BINARY_OP_MULTIPLY_FLOAT               133
#define BINARY_OP_MULTIPLY_INT                 134
#define BINARY_OP_SUBSCR_DICT                  135
#define BINARY_OP_SUBSCR_GETITEM               136
#define BINARY_OP_SUBSCR_LIST_INT              137
#define BINARY_OP_SUBSCR_LIST_SLICE            138
#define BINARY_OP_SUBSCR_STR_INT               139
#define BINARY_OP_SUBSCR_TUPLE_INT             140
#define BINARY_OP_SUBTRACT_FLOAT               141
#define BINARY_OP_SUBTRACT_INT                 142
#define CALL_ALLOC_AND_ENTER_INIT              143
#define CALL_BOUND_METHOD_EXACT_ARGS           144
#define CALL_BOUND_METHOD_GENERAL              145
#define CALL_BUILTIN_CLASS                     146
#define CALL_BUILTIN_FAST                      147
#define CALL_BUILTIN_FAST_WITH_KEYWORDS        148
#define CALL_BUILTIN_O                         149
#define CALL_ISINSTANCE                        150
#define CALL_KW_BOUND_METHOD                   151
#define CALL_KW_NON_PY                         152
#define CALL_KW_PY                             153
#define CALL_LEN                               154
#define CALL_LIST_APPEND                       155
#define CALL_METHOD_DESCRIPTOR_FAST            156
#define CALL_METHOD_DESCRIPTOR_FAST_WITH_KEYWORDS 157
#define CALL_METHOD_DESCRIPTOR_NOARGS          158
#define CALL_METHOD_DESCRIPTOR_O               159
#define CALL_NON_PY_GENERAL                    160
#define CALL_PY_EXACT_ARGS                     161
#define CALL_PY_GENERAL                        162
#define CALL_STR_1                             163
#define CALL_TUPLE_1                           164
#define CALL_TYPE_1                            165
#define COMPARE_OP_FLOAT                       166
#define COMPARE_OP_INT                         167
#define COMPARE_OP_STR                         168
#define CONTAINS_OP_DICT                       169
#define CONTAINS_OP_SET                        170
#define FOR_ITER_GEN                           171
#define FOR_ITER_LIST                          172
#define FOR_ITER_RANGE                         173
#define FOR_ITER_TUPLE                         174
#define JUMP_BACKWARD_JIT                      175
#define JUMP_BACKWARD_NO_JIT                   176
#define LOAD_ATTR_CLASS                        177
#define LOAD_ATTR_CLASS_WITH_METACLASS_CHECK   178
#define LOAD_ATTR_GETATTRIBUTE_OVERRIDDEN      179
#define LOAD_ATTR_INSTANCE_VALUE               180
#define LOAD_ATTR_METHOD_LAZY_DICT             181
#define LOAD_ATTR_METHOD_NO_DICT               182
#define LOAD_ATTR_METHOD_WITH_VALUES           183
#define LOAD_ATTR_MODULE                       184
#define LOAD_ATTR_NONDESCRIPTOR_NO_DICT        185
#define LOAD_ATTR_NONDESCRIPTOR_WITH_VALUES    186
#define LOAD_ATTR_PROPERTY                     187
#define LOAD_ATTR_SLOT                         188
#define LOAD_ATTR_WITH_HINT                    189
#define LOAD_CONST_IMMORTAL                    190
#define LOAD_CONST_MORTAL                      191
#define LOAD_GLOBAL_BUILTIN                    192
#define LOAD_GLOBAL_MODULE                     193
#define LOAD_SUPER_ATTR_ATTR                   194
#define LOAD_SUPER_ATTR_METHOD                 195
#define RESUME_CHECK                           196
#define SEND_GEN                               197
#define STORE_ATTR_INSTANCE_VALUE              198
#define STORE_ATTR_SLOT                        199
#define STORE_ATTR_WITH_HINT                   200
#define STORE_SUBSCR_DICT                      201
#define STORE_SUBSCR_LIST_INT                  202
#define TO_BOOL_ALWAYS_TRUE                    203
#define TO_BOOL_BOOL                           204
#define TO_BOOL_INT                            205
#define TO_BOOL_LIST                           206
#define TO_BOOL_NONE                           207
#define TO_BOOL_STR                            208
#define UNPACK_SEQUENCE_LIST                   209
#define UNPACK_SEQUENCE_TUPLE                  210
#define UNPACK_SEQUENCE_TWO_TUPLE              211
#define INSTRUMENTED_END_FOR                   234
#define INSTRUMENTED_POP_ITER                  235
#define INSTRUMENTED_END_SEND                  236
#define INSTRUMENTED_FOR_ITER                  237
#define INSTRUMENTED_INSTRUCTION               238
#define INSTRUMENTED_JUMP_FORWARD              239
#define INSTRUMENTED_NOT_TAKEN                 240
#define INSTRUMENTED_POP_JUMP_IF_TRUE          241
#define INSTRUMENTED_POP_JUMP_IF_FALSE         242
#define INSTRUMENTED_POP_JUMP_IF_NONE          243
#define INSTRUMENTED_POP_JUMP_IF_NOT_NONE      244
#define INSTRUMENTED_RESUME                    245
#define INSTRUMENTED_RETURN_VALUE              246
#define INSTRUMENTED_YIELD_VALUE               247
#define INSTRUMENTED_END_ASYNC_FOR             248
#define INSTRUMENTED_LOAD_SUPER_ATTR           249
#define INSTRUMENTED_CALL                      250
#define INSTRUMENTED_CALL_KW                   251
#define INSTRUMENTED_CALL_FUNCTION_EX          252
#define INSTRUMENTED_JUMP_BACKWARD             253
#define INSTRUMENTED_LINE                      254
#define ENTER_EXECUTOR                         255
#define ANNOTATIONS_PLACEHOLDER                256
#define JUMP                                   257
#define JUMP_IF_FALSE                          258
#define JUMP_IF_TRUE                           259
#define JUMP_NO_INTERRUPT                      260
#define LOAD_CLOSURE                           261
#define POP_BLOCK                              262
#define SETUP_CLEANUP                          263
#define SETUP_FINALLY                          264
#define SETUP_WITH                             265
#define STORE_FAST_MAYBE_NULL                  266

#define HAVE_ARGUMENT                           43
#define MIN_SPECIALIZED_OPCODE                 129
#define MIN_INSTRUMENTED_OPCODE                234

#ifdef __cplusplus
}
#endif
#endif /* !Py_OPCODE_IDS_H */
